from calendar import timegm
from datetime import datetime, timezone

from django.conf import settings
from django.utils.functional import lazy
from django.utils.timezone import is_naive, make_aware


def make_utc(dt):
    if settings.USE_TZ and is_naive(dt):
        return make_aware(dt, timezone=timezone.utc)

    return dt


def aware_utcnow():
    return make_utc(datetime.utcnow())


def datetime_to_epoch(dt):
    return timegm(dt.utctimetuple())


def datetime_from_epoch(ts):
    return make_utc(datetime.utcfromtimestamp(ts))


def format_lazy(s, *args, **kwargs):
    return s.format(*args, **kwargs)


format_lazy = lazy(format_lazy, str)



from django.utils.translation import gettext_lazy as _
from rest_framework import exceptions, status


class TokenError(Exception):
    pass


class TokenBackendError(Exception):
    pass


class DetailDictMixin:
    def __init__(self, detail=None, code=None):
        """
        Builds a detail dictionary for the error to give more information to API
        users.
        """
        detail_dict = {"detail": self.default_detail, "code": self.default_code}

        if isinstance(detail, dict):
            detail_dict.update(detail)
        elif detail is not None:
            detail_dict["detail"] = detail

        if code is not None:
            detail_dict["code"] = code

        super().__init__(detail_dict)


class AuthenticationFailed(DetailDictMixin, exceptions.AuthenticationFailed):
    pass


class InvalidToken(AuthenticationFailed):
    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = _("Token is invalid or expired")
    default_code = "token_not_valid"



import datetime
import time
from typing import List, Union

from django.core.cache import cache
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from nilva_django_auth.models import UserSession
from nilva_django_auth.simplejwt.exceptions import TokenError, InvalidToken
from nilva_django_auth.settings import api_settings
from nilva_django_auth.simplejwt.utils import datetime_from_epoch


class SessionManagement:
    @classmethod
    def blacklist(cls, user_id, jti):
        cls.check_blacklist(jti)
        UserSession.objects.filter(jti=jti, user_id=user_id).update(is_active=False)
        cls._invalidate_cache(jti)

    @classmethod
    def check_blacklist(cls, jti, raise_exception=True):
        cache_key = cls._blacklist_cache_key(jti)
        if cache_data := cache.get(cache_key):
            if cache_data and raise_exception:
                raise TokenError(_("Token is blacklisted"))
            return cache_data

        session = UserSession.objects.filter(jti=jti, is_active=True, exp__gte=timezone.now()).first()
        blacklisted = not session

        cache.set(cache_key, blacklisted, api_settings.BLACKLIST_CACHE_TIMEOUT)
        if blacklisted and raise_exception:
            raise TokenError(_("Token is blacklisted"))

        return blacklisted

    @classmethod
    def _blacklist_cache_key(cls, jti):
        return f"SessionManagement:token_blacklist_{jti}"

    @classmethod
    def _invalidate_cache(cls, jti: Union[str, List[str]]):
        if isinstance(jti, str):
            jti = [jti]
        cache.delete_many([cls._blacklist_cache_key(j) for j in jti])

    @classmethod
    def _invalidate_user(cls, user_id):
        cache.delete(cls.get_session_list_cache_key(user_id))

    @classmethod
    def blacklist_except(cls, user_id, exclude_jti, min_session_age=None):
        """
        Blacklist all sessions for a user except the one with the specified JTI.
        Only sessions created before the current session will be blacklisted.

        Args:
            user_id: The ID of the user whose sessions should be blacklisted
            exclude_jti: The JTI of the session to exclude from blacklisting
            min_session_age: Deprecated, kept for backward compatibility
        """
        # Get the current session to use its creation time as reference
        cls.check_session_age(exclude_jti, user_id, min_session_age)

        # Get all active sessions for this user except the current one
        # that were created before the current session
        sessions_query = UserSession.objects.filter(
            user_id=user_id,
            is_active=True,
        ).exclude(jti=exclude_jti)

        jti_list = list(sessions_query.values_list('jti', flat=True))
        if jti_list:
            cls._invalidate_cache(jti_list)
        sessions_query.update(is_active=False)

    @classmethod
    def check_session_age(cls, jti, user_id, min_session_age=None):
        if not min_session_age:
            min_session_age = api_settings.BLACKLIST_ALL_MIN_SESSION_AGE
        current_session = UserSession.objects.filter(user_id=user_id, jti=jti, is_active=True).first()
        if not current_session:
            # If current session not found, don't blacklist anything
            raise InvalidToken(_("Token is blacklisted or expired"))
        elif current_session.created_at > timezone.now() - min_session_age:
            # If the current session is too new, don't blacklist anything
            raise InvalidToken(
                _("Current session is too new. Please wait at least {} minutes before blacklisting other sessions.").format(
                    int(min_session_age.total_seconds() / 60)))

    @classmethod
    def create_session(cls, user, jti=None, exp=None, request=None):
        """
        Create a new session for the user.

        Args:
            user: The user for whom the session is being created
            jti: The JWT ID (optional)
            exp: The expiration time (optional, defaults to 24 hours from now)
                Can be a datetime object or a Unix timestamp (float)
            request: The request object (optional, used to extract device info and IP address)

        Returns:
            UserSession: The created session object
        """
        cls._invalidate_user(user.id)
        # Convert exp from timestamp to datetime if it's not already a datetime
        if exp is not None and not isinstance(exp, datetime.datetime):
            exp = datetime_from_epoch(exp)
        return UserSession.create_session(
            user=user,
            jti=jti,
            exp=exp,
            request=request
        )

    @classmethod
    def update_session(cls, old_jti, jti, exp, user_id):
        cls._invalidate_user(user_id)
        # Convert exp from timestamp to datetime if it's not already a datetime
        if not isinstance(exp, datetime.datetime):
            exp = datetime_from_epoch(exp)
        UserSession.objects.filter(jti=old_jti).update(
            exp=exp, jti=jti, last_activity=timezone.now(), is_active=True
        )

    @classmethod
    def deactivate_expired_sessions(cls, user_id=None, check_time=None):
        """
        Set is_active to False for all expired sessions.

        Args:
            user_id (int, optional): If provided, only deactivate expired sessions for this user.
            check_time (datetime or float, optional): The time to check against. If not provided, uses current time.
                If a float is provided, it's treated as a Unix timestamp and converted to datetime.

        Returns:
            int: The number of sessions that were deactivated.
        """
        if check_time is None:
            check_time = timezone.now()
        elif not isinstance(check_time, datetime.datetime):
            # Convert from Unix timestamp to datetime if necessary
            check_time = datetime_from_epoch(check_time)

        # Build the query to find expired sessions that are still active
        query = UserSession.objects.filter(is_active=True, exp__lt=check_time)

        # If user_id is provided, filter by user_id
        if user_id is not None:
            query = query.filter(user_id=user_id)

        # Update the sessions to set is_active to False
        count = query.update(is_active=False)

        return count

    @classmethod
    def get_session_list_cache_key(cls, user_id):
        return f"SessionManagement:session_list_{user_id}"

    @classmethod
    def terminate_sessions(cls, user_id, session_ids, current_jti=None, min_session_age=None):
        """
        Terminate specific sessions for a user.

        Args:
            user_id: The ID of the user whose sessions should be terminated
            session_ids: List of session IDs to terminate
            current_jti: The JTI of the current session (optional)
            min_session_age: Minimum age required for the current session (optional)

        Returns:
            int: The number of sessions that were terminated
        """
        cls.check_session_age(current_jti, user_id, min_session_age)

        # Get the sessions to terminate
        sessions = UserSession.objects.filter(
            user_id=user_id,
            id__in=session_ids,
            is_active=True
        )

        # Get the JTIs of the sessions to terminate
        jti_list = list(sessions.values_list('jti', flat=True))

        # Blacklist the sessions
        if jti_list:
            # Invalidate cache for the JTIs
            cls._invalidate_cache(jti_list)
            # Deactivate the sessions
            sessions.update(is_active=False)

        return len(jti_list)



from rest_framework import serializers

from nilva_django_auth.models import UserSession


class SessionSerializer(serializers.ModelSerializer):
    """
    Serializer for user sessions.
    """
    os = serializers.CharField(read_only=True)
    device_brand = serializers.CharField(read_only=True)
    device_model = serializers.CharField(read_only=True)
    device_family = serializers.CharField(read_only=True)
    is_mobile = serializers.BooleanField(read_only=True)
    browser = serializers.CharField(read_only=True)
    last_online = serializers.SerializerMethodField()
    is_current = serializers.SerializerMethodField()

    class Meta:
        model = UserSession
        fields = [
            'user_agent',
            'ip_address',
            'created_at',
            'os',
            'device_brand',
            'device_model',
            'device_family',
            'is_mobile',
            'browser',
            'last_online',
            'is_current',
        ]
        read_only_fields = fields

    def get_last_online(self, obj):
        return obj.last_activity.isoformat()

    def get_is_current(self, obj: UserSession):
        current_jti = self.context.get('current_jti')
        jti = str(obj.jti)

        return jti == current_jti



# Generated by Django 5.2.4 on 2025-08-08 12:34

import uuid
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="UserSession",
            fields=[
                (
                    "id",
                    models.CharField(default=uuid.uuid4, max_length=36, primary_key=True, serialize=False),
                ),
                (
                    "jti",
                    models.UUIDField(unique=True, verbose_name="JWT ID"),
                ),
                (
                    "user_agent",
                    models.TextField(blank=True, null=True, verbose_name="User Agent"),
                ),
                (
                    "ip_address",
                    models.GenericIPAddressField(
                        blank=True, null=True, verbose_name="IP Address"
                    ),
                ),
                (
                    "referer",
                    models.URLField(blank=True, null=True, verbose_name="Referer"),
                ),
                ("exp", models.DateTimeField(verbose_name="Expiration Time")),
                (
                    "created_at",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="Created At"
                    ),
                ),
                (
                    "last_activity",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="Last Activity"
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(default=True, verbose_name="Is Active"),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="sessions",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="User",
                    ),
                ),
            ],
            options={
                "verbose_name": "User Session",
                "verbose_name_plural": "User Sessions",
                "ordering": ["-last_activity"],
            },
        ),
    ]



from datetime import timedelta
from django.conf import settings
from django.test.signals import setting_changed
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

def format_lazy(format_string, *args, **kwargs):
    """
    Lazily format a lazy string.
    """
    return format_string.format(*args, **kwargs) if args or kwargs else format_string

DEFAULTS = {
    # Cookie settings
    'USE_COOKIES': False,
    'AUTH_COOKIE_NAME': 'auth',
    'REFRESH_COOKIE_NAME': 'refresh',
    'SECURE_COOKIES': True,
    'HTTPONLY_COOKIES': True,
    'SAMESITE_COOKIES': 'Lax',

    # Custom user model
    'USER_MODEL': 'auth.User',

    # Session management
    'BLACKLIST_ALL_MIN_SESSION_AGE': timedelta(minutes=5),

    # SimpleJWT settings
    "ACCESS_TOKEN_LIFETIME": timedelta(minutes=5),
    "REFRESH_TOKEN_LIFETIME": timedelta(days=1),
    "ROTATE_REFRESH_TOKENS": False,
    "BLACKLIST_AFTER_ROTATION": False,
    "UPDATE_LAST_LOGIN": False,
    "BLACKLIST_CACHE_TIMEOUT": 300,
    "ALGORITHM": "HS256",
    "SIGNING_KEY": settings.SECRET_KEY,
    "VERIFYING_KEY": "",
    "AUDIENCE": None,
    "ISSUER": None,
    "JSON_ENCODER": None,
    "JWK_URL": None,
    "LEEWAY": 0,
    "AUTH_HEADER_TYPES": ("Bearer",),
    "AUTH_HEADER_NAME": "HTTP_AUTHORIZATION",
    "USER_ID_FIELD": "id",
    "USER_ID_CLAIM": "user_id",
    "USER_AUTHENTICATION_RULE": "nilva_django_auth.simplejwt.authentication.default_user_authentication_rule",
    "AUTH_TOKEN_CLASSES": ("nilva_django_auth.simplejwt.tokens.AccessToken",),
    "TOKEN_TYPE_CLAIM": "token_type",
    "JTI_CLAIM": "jti",
    "TOKEN_USER_CLASS": "nilva_django_auth.simplejwt.models.TokenUser",
    "SLIDING_TOKEN_REFRESH_EXP_CLAIM": "refresh_exp",
    "SLIDING_TOKEN_LIFETIME": timedelta(minutes=5),
    "SLIDING_TOKEN_REFRESH_LIFETIME": timedelta(days=1),
    "TOKEN_OBTAIN_SERIALIZER": "nilva_django_auth.simplejwt.serializers.TokenObtainPairSerializer",
    "TOKEN_REFRESH_SERIALIZER": "nilva_django_auth.simplejwt.serializers.TokenRefreshSerializer",
    "TOKEN_BLACKLIST_SERIALIZER": "nilva_django_auth.simplejwt.serializers.TokenBlacklistSerializer",
    "TOKEN_BLACKLIST_ALL_SERIALIZER": "nilva_django_auth.simplejwt.serializers.TokenBlacklistAllSerializer",
    "TOKEN_BLACKLIST_SESSIONS_SERIALIZER": "nilva_django_auth.simplejwt.serializers.TokenBlacklistSessionsSerializer",
    "TOKEN_VERIFY_SERIALIZER": "nilva_django_auth.simplejwt.serializers.TokenVerifySerializer",
    "SLIDING_TOKEN_OBTAIN_SERIALIZER": "nilva_django_auth.simplejwt.serializers.TokenObtainSlidingSerializer",
    "SLIDING_TOKEN_REFRESH_SERIALIZER": "nilva_django_auth.simplejwt.serializers.TokenRefreshSlidingSerializer",
}

IMPORT_STRINGS = [
    # SimpleJWT import strings
    "AUTH_TOKEN_CLASSES",
    "JSON_ENCODER",
    "TOKEN_USER_CLASS",
    "USER_AUTHENTICATION_RULE",
]

REMOVED_SETTINGS = [
    # SimpleJWT removed settings
    "AUTH_HEADER_TYPE",
    "AUTH_TOKEN_CLASS",
    "SECRET_KEY",
    "TOKEN_BACKEND_CLASS",
]


def perform_import(val, setting_name):
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if val is None:
        return None
    elif isinstance(val, str):
        return import_from_string(val, setting_name)
    elif isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val


def import_from_string(val, setting_name):
    """
    Attempt to import a class from a string representation.
    """
    try:
        return import_string(val)
    except ImportError as e:
        msg = "Could not import '%s' for API setting '%s'. %s: %s." % (
            val, setting_name, e.__class__.__name__, e)
        raise ImportError(msg)


class APISettings:
    """
    A settings object that allows accessing the NILVA_AUTH settings as properties.
    For example:
        from nilva_django_auth.settings import api_settings
        print(api_settings.TOKEN_LIFETIME)
    """
    def __init__(self, user_settings=None, defaults=None, import_strings=None, removed_settings=None):
        if user_settings:
            self._user_settings = user_settings
        self.defaults = defaults or DEFAULTS
        self.import_strings = import_strings or IMPORT_STRINGS
        self.removed_settings = removed_settings or REMOVED_SETTINGS
        self._cached_attrs = set()

    @property
    def user_settings(self):
        if not hasattr(self, '_user_settings'):
            self._user_settings = {}
            # Get settings from NILVA_AUTH
            nilva_auth_settings = getattr(settings, 'NILVA_AUTH', {})
            self._user_settings.update(nilva_auth_settings)

            # For backward compatibility, if SIMPLE_JWT is defined, use it for SimpleJWT-specific settings
            simple_jwt_settings = getattr(settings, 'SIMPLE_JWT', {})
            if simple_jwt_settings:
                # Only update with SimpleJWT settings if they don't already exist in NILVA_AUTH
                for key, value in simple_jwt_settings.items():
                    if key not in self._user_settings:
                        self._user_settings[key] = value

            # Check for removed settings
            self.__check_user_settings(self._user_settings)
        return self._user_settings

    def __check_user_settings(self, user_settings):
        for setting in self.removed_settings:
            if setting in user_settings:
                raise RuntimeError(
                    format_lazy(
                        _(
                            "The '{}' setting has been removed. Please refer to documentation for available settings."
                        ),
                        setting,
                    )
                )
        return user_settings

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError("Invalid API setting: '%s'" % attr)

        try:
            # Check if present in user settings
            val = self.user_settings[attr]
        except KeyError:
            # Fall back to defaults
            val = self.defaults[attr]

        # Coerce import strings into classes
        if attr in self.import_strings:
            val = perform_import(val, attr)

        # Cache the result
        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def reload(self):
        """
        Reload settings after a change. Used mainly in testing.
        """
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, '_user_settings'):
            delattr(self, '_user_settings')


api_settings = APISettings(None, DEFAULTS, IMPORT_STRINGS, REMOVED_SETTINGS)


def reload_api_settings(*args, **kwargs):
    """
    Reset settings. Used in testing.
    """
    setting = kwargs['setting']
    if setting in ['NILVA_AUTH', 'SIMPLE_JWT']:
        api_settings.reload()


setting_changed.connect(reload_api_settings)



from django.apps import AppConfig


class NilvaDjangoAuthConfig(AppConfig):
    name = 'nilva_django_auth'
    verbose_name = 'Nilva Django Authentication'
    
    def ready(self):
        # Import signal handlers or perform other initialization
        pass



import datetime
import uuid

from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from user_agents.parsers import parse

from nilva_django_auth.simplejwt.utils import datetime_from_epoch


class UserSession(models.Model):
    """
    Model to track user sessions.
    """
    id = models.CharField(max_length=36, default=uuid.uuid4, primary_key=True)

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name='sessions',
        verbose_name=_('User')
    )
    jti = models.UUIDField(
        unique=True,
        verbose_name=_('JWT ID')
    )
    user_agent = models.TextField(
        blank=True,
        null=True,
        verbose_name=_('User Agent')
    )
    ip_address = models.GenericIPAddressField(
        blank=True,
        null=True,
        verbose_name=_('IP Address')
    )
    referer = models.URLField(
        blank=True,
        null=True,
        verbose_name=_('Referer')
    )
    exp = models.DateTimeField(
        verbose_name=_('Expiration Time')
    )
    created_at = models.DateTimeField(
        default=timezone.now,
        verbose_name=_('Created At')
    )
    last_activity = models.DateTimeField(
        default=timezone.now,
        verbose_name=_('Last Activity')
    )
    is_active = models.BooleanField(
        default=True,
        verbose_name=_('Is Active')
    )

    class Meta:
        verbose_name = _('User Session')
        verbose_name_plural = _('User Sessions')
        ordering = ['-last_activity']

    def __str__(self):
        return f"{self.user} - {self.jti}"

    def update_activity(self):
        """
        Update the last activity timestamp.
        """
        self.last_activity = timezone.now()
        self.save(update_fields=['last_activity'])

    def deactivate(self):
        """
        Deactivate the session.
        """
        self.is_active = False
        self.save(update_fields=['is_active'])

    @classmethod
    def create_session(cls, user, jti, exp, request=None):
        """
        Create a new session for the user.

        Args:
            user: The user for whom the session is being created
            jti: The JWT ID
            exp: The expiration time (datetime object or Unix timestamp)
            request: The request object (optional, used to extract device info and IP address)

        Returns:
            UserSession: The created session object
        """
        ip_address = None
        user_agent_str = None
        referer = None

        # Convert exp from timestamp to datetime if it's not already a datetime
        if exp is not None and not isinstance(exp, datetime.datetime):
            exp = datetime_from_epoch(exp)
        # If exp is None, set a default expiration time (24 hours from now)
        elif exp is None:
            exp = timezone.now() + datetime.timedelta(hours=24)

        if request:
            # Extract device info from user agent
            user_agent_str = request.META.get('HTTP_USER_AGENT', '')

            # We don't need to parse user agent here as we'll compute properties on-the-fly when needed

            # Extract IP address
            x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
            if x_forwarded_for:
                ip_address = x_forwarded_for.split(',')[0]
            else:
                ip_address = request.META.get('REMOTE_ADDR')

            # Extract referer
            referer = request.META.get('HTTP_REFERER')

        return cls.objects.create(
            user=user,
            jti=jti,
            user_agent=user_agent_str,
            ip_address=ip_address,
            referer=referer,
            exp=exp
        )

    @classmethod
    def deactivate_all_for_user(cls, user):
        """
        Deactivate all sessions for the user.
        """
        cls.objects.filter(user=user, is_active=True).update(is_active=False)

    @cached_property
    def user_agent_obj(self):
        """
        Parse and return the user agent object.
        """
        return self.user_agent and parse(self.user_agent)

    @property
    def os(self):
        """
        Get the operating system family from the user agent.
        """
        return self.user_agent_obj and self.user_agent_obj.os.family

    @property
    def device_brand(self):
        """
        Get the device brand from the user agent.
        """
        return self.user_agent_obj and self.user_agent_obj.device.brand

    @property
    def device_model(self):
        """
        Get the device model from the user agent.
        """
        return self.user_agent_obj and self.user_agent_obj.device.model

    @property
    def device_family(self):
        """
        Get the device family from the user agent.
        """
        return self.user_agent_obj and self.user_agent_obj.device.family

    @property
    def is_mobile(self):
        """
        Check if the device is mobile from the user agent.
        """
        return self.user_agent_obj and self.user_agent_obj.is_mobile

    @property
    def browser(self):
        """
        Get the browser family from the user agent.
        """
        return self.user_agent_obj and self.user_agent_obj.browser.family



from django.contrib.auth import authenticate, get_user_model
from django.contrib.auth.models import update_last_login
from django.utils.translation import gettext_lazy as _
from rest_framework import exceptions, serializers

from .settings import api_settings
from .tokens import RefreshToken, SlidingToken, UntypedToken
from nilva_django_auth.session_management import SessionManagement


class PasswordField(serializers.CharField):
    def __init__(self, *args, **kwargs):
        kwargs.setdefault("style", {})

        kwargs["style"]["input_type"] = "password"
        kwargs["write_only"] = True

        super().__init__(*args, **kwargs)


class TokenObtainSerializer(serializers.Serializer):
    username_field = get_user_model().USERNAME_FIELD
    token_class = None

    default_error_messages = {
        "no_active_account": _("No active account found with the given credentials")
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields[self.username_field] = serializers.CharField()
        self.fields["password"] = PasswordField()

    def validate(self, attrs):
        authenticate_kwargs = {
            self.username_field: attrs[self.username_field],
            "password": attrs["password"],
        }
        try:
            authenticate_kwargs["request"] = self.context["request"]
        except KeyError:
            pass

        self.user = authenticate(**authenticate_kwargs)

        if not api_settings.USER_AUTHENTICATION_RULE(self.user):
            raise exceptions.AuthenticationFailed(
                self.error_messages["no_active_account"],
                "no_active_account",
            )

        return {}

    @classmethod
    def get_token(cls, user):
        return cls.token_class.for_user(user)


class TokenObtainPairSerializer(TokenObtainSerializer):
    token_class = RefreshToken


    def validate(self, attrs):
        data = super().validate(attrs)

        refresh = self.get_token(self.user)

        data["refresh"] = str(refresh)
        data["access"] = str(refresh.access_token)

        self.create_session(refresh)

        if api_settings.UPDATE_LAST_LOGIN:
            update_last_login(None, self.user)

        return data

    def create_session(self, refresh):
        jti = str(refresh["jti"])
        exp = refresh["exp"]
        SessionManagement.create_session(
            user=self.user,
            jti=jti,
            exp=exp,
            request=self.context.get("request")
        )


class TokenObtainSlidingSerializer(TokenObtainSerializer):
    token_class = SlidingToken

    def validate(self, attrs):
        data = super().validate(attrs)

        token = self.get_token(self.user)

        data["token"] = str(token)

        if api_settings.UPDATE_LAST_LOGIN:
            update_last_login(None, self.user)

        return data


class TokenRefreshSerializer(serializers.Serializer):
    refresh = serializers.CharField()
    access = serializers.CharField(read_only=True)
    token_class = RefreshToken

    def validate(self, attrs):
        refresh = self.token_class(attrs["refresh"])
        old_jti = refresh[api_settings.JTI_CLAIM]

        data = {"access": str(refresh.access_token)}

        if api_settings.ROTATE_REFRESH_TOKENS:
            if api_settings.BLACKLIST_AFTER_ROTATION:
                try:
                    # Attempt to blacklist the given refresh token
                    refresh.check_blacklist()
                except AttributeError:
                    # If blacklist app not installed, `blacklist` method will
                    # not be present
                    pass

            refresh.set_jti()
            refresh.set_exp()
            refresh.set_iat()

            data["refresh"] = str(refresh)

        if jti := refresh.get(api_settings.JTI_CLAIM) != old_jti:
            exp = refresh["exp"]
            user_id = refresh[api_settings.USER_ID_CLAIM]
            SessionManagement.update_session(old_jti, jti, exp, user_id)

        return data


class TokenRefreshSlidingSerializer(serializers.Serializer):
    token = serializers.CharField()
    token_class = SlidingToken

    def validate(self, attrs):
        token = self.token_class(attrs["token"])

        # Check that the timestamp in the "refresh_exp" claim has not
        # passed
        token.check_exp(api_settings.SLIDING_TOKEN_REFRESH_EXP_CLAIM)

        # Update the "exp" and "iat" claims
        token.set_exp()
        token.set_iat()

        return {"token": str(token)}


class TokenVerifySerializer(serializers.Serializer):
    token = serializers.CharField()

    def validate(self, attrs):
        token = UntypedToken(attrs["token"])

        if api_settings.BLACKLIST_AFTER_ROTATION:
            jti = token.get(api_settings.JTI_CLAIM)
            SessionManagement.check_blacklist(jti)

        return {}


class TokenBlacklistSerializer(serializers.Serializer):
    refresh = serializers.CharField()
    token_class = RefreshToken

    def validate(self, attrs):
        refresh = self.token_class(attrs["refresh"])
        try:
            refresh.blacklist()
        except AttributeError:
            pass
        return {}


class TokenBlacklistAllSerializer(serializers.Serializer):
    refresh = serializers.CharField()
    token_class = RefreshToken

    def validate(self, attrs):
        refresh = self.token_class(attrs["refresh"])
        try:
            refresh.blacklist_except()
        except AttributeError:
            pass
        return {}


class TokenBlacklistSessionsSerializer(serializers.Serializer):
    refresh = serializers.CharField()
    session_ids = serializers.ListField(
        child=serializers.CharField(),
        required=True,
        help_text=_("List of session IDs to terminate")
    )
    token_class = RefreshToken

    def validate(self, attrs):
        refresh = self.token_class(attrs["refresh"])
        user_id = refresh.get(api_settings.USER_ID_CLAIM)
        jti = refresh.get(api_settings.JTI_CLAIM)
        session_ids = attrs["session_ids"]

        # Use SessionManagement to terminate the sessions
        terminated_count = SessionManagement.terminate_sessions(
            user_id=user_id,
            session_ids=session_ids,
            current_jti=jti
        )

        return {"terminated_sessions": terminated_count}



from django.contrib.auth import models as auth_models
from django.db.models.manager import EmptyManager
from django.utils.functional import cached_property

from .compat import CallableFalse, CallableTrue
from .settings import api_settings


class TokenUser:
    """
    A dummy user class modeled after django.contrib.auth.models.AnonymousUser.
    Used in conjunction with the `JWTStatelessUserAuthentication` backend to
    implement single sign-on functionality across services which share the same
    secret key.  `JWTStatelessUserAuthentication` will return an instance of this
    class instead of a `User` model instance.  Instances of this class act as
    stateless user objects which are backed by validated tokens.
    """

    # User is always active since Simple JWT will never issue a token for an
    # inactive user
    is_active = True

    _groups = EmptyManager(auth_models.Group)
    _user_permissions = EmptyManager(auth_models.Permission)

    def __init__(self, token):
        self.token = token

    def __str__(self):
        return f"TokenUser {self.id}"

    @cached_property
    def id(self):
        return self.token[api_settings.USER_ID_CLAIM]

    @cached_property
    def pk(self):
        return self.id

    @cached_property
    def username(self):
        return self.token.get("username", "")

    @cached_property
    def is_staff(self):
        return self.token.get("is_staff", False)

    @cached_property
    def is_superuser(self):
        return self.token.get("is_superuser", False)

    def __eq__(self, other):
        return self.id == other.id

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.id)

    def save(self):
        raise NotImplementedError("Token users have no DB representation")

    def delete(self):
        raise NotImplementedError("Token users have no DB representation")

    def set_password(self, raw_password):
        raise NotImplementedError("Token users have no DB representation")

    def check_password(self, raw_password):
        raise NotImplementedError("Token users have no DB representation")

    @property
    def groups(self):
        return self._groups

    @property
    def user_permissions(self):
        return self._user_permissions

    def get_group_permissions(self, obj=None):
        return set()

    def get_all_permissions(self, obj=None):
        return set()

    def has_perm(self, perm, obj=None):
        return False

    def has_perms(self, perm_list, obj=None):
        return False

    def has_module_perms(self, module):
        return False

    @property
    def is_anonymous(self):
        return CallableFalse

    @property
    def is_authenticated(self):
        return CallableTrue

    def get_username(self):
        return self.username

    def __getattr__(self, attr):
        """This acts as a backup attribute getter for custom claims defined in Token serializers."""
        return self.token.get(attr, None)



from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from .models import UserSession


@admin.register(UserSession)
class UserSessionAdmin(admin.ModelAdmin):
    list_display = ('user', 'ip_address', 'last_activity', 'is_active', 'os', 'browser')
    list_filter = ('is_active', 'created_at', 'last_activity')
    search_fields = ('user__username', 'user__email', 'ip_address')
    readonly_fields = ('id', 'user', 'jti', 'user_agent', 'ip_address',
                      'exp', 'created_at', 'last_activity', 'os', 'browser', 
                      'device_brand', 'device_model', 'device_family', 'is_mobile', 'referer')
    
    fieldsets = (
        (_('User Information'), {
            'fields': ('user',)
        }),
        (_('Session Information'), {
            'fields': ('id', 'jti', 'exp', 'is_active', 'created_at', 'last_activity', 'referer')
        }),
        (_('Device Information'), {
            'fields': ('user_agent', 'os', 'browser', 'device_brand',
                      'device_model', 'device_family', 'is_mobile')
        }),
        (_('Location Information'), {
            'fields': ('ip_address',)
        }),
    )
    
    def has_add_permission(self, request):
        # Sessions should only be created through the API, not manually
        return False
    
    def has_change_permission(self, request, obj=None):
        # Only allow changing the is_active status
        return True
    
    def get_readonly_fields(self, request, obj=None):
        # All fields are readonly except is_active
        if obj:
            return [f.name for f in obj._meta.fields if f.name != 'is_active']
        return self.readonly_fields


# This file is kept for backward compatibility
# All settings have been moved to nilva_django_auth/settings.py

from nilva_django_auth.settings import api_settings

# Re-export api_settings for backward compatibility
# This ensures that existing imports continue to work



from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _
from rest_framework import HTTP_HEADER_ENCODING, authentication

from .exceptions import AuthenticationFailed, InvalidToken, TokenError
from .settings import api_settings
from nilva_django_auth.session_management import SessionManagement

AUTH_HEADER_TYPES = api_settings.AUTH_HEADER_TYPES

if not isinstance(api_settings.AUTH_HEADER_TYPES, (list, tuple)):
    AUTH_HEADER_TYPES = (AUTH_HEADER_TYPES,)

AUTH_HEADER_TYPE_BYTES = {h.encode(HTTP_HEADER_ENCODING) for h in AUTH_HEADER_TYPES}


class JWTAuthentication(authentication.BaseAuthentication):
    """
    An authentication plugin that authenticates requests through a JSON web
    token provided in a request header.
    """

    www_authenticate_realm = "api"
    media_type = "application/json"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.user_model = get_user_model()

    def authenticate(self, request):
        header = self.get_header(request)
        if header is None:
            return None

        raw_token = self.get_raw_token(header)
        if raw_token is None:
            return None

        validated_token = self.get_validated_token(raw_token)

        return self.get_user(validated_token), validated_token

    def authenticate_header(self, request):
        return '{} realm="{}"'.format(
            AUTH_HEADER_TYPES[0],
            self.www_authenticate_realm,
        )

    def get_header(self, request):
        """
        Extracts the header containing the JSON web token from the given
        request.
        """
        header = request.META.get(api_settings.AUTH_HEADER_NAME)

        if isinstance(header, str):
            # Work around django test client oddness
            header = header.encode(HTTP_HEADER_ENCODING)

        return header

    def get_raw_token(self, header):
        """
        Extracts an unvalidated JSON web token from the given "Authorization"
        header value.
        """
        parts = header.split()

        if len(parts) == 0:
            # Empty AUTHORIZATION header sent
            return None

        if parts[0] not in AUTH_HEADER_TYPE_BYTES:
            # Assume the header does not contain a JSON web token
            return None

        if len(parts) != 2:
            raise AuthenticationFailed(
                _("Authorization header must contain two space-delimited values"),
                code="bad_authorization_header",
            )

        return parts[1]

    def get_validated_token(self, raw_token):
        """
        Validates an encoded JSON web token and returns a validated token
        wrapper object.
        """
        messages = []
        for AuthToken in api_settings.AUTH_TOKEN_CLASSES:
            try:
                token = AuthToken(raw_token)
                self.check_blacklist(token)

                return token
            except TokenError as e:
                messages.append(
                    {
                        "token_class": AuthToken.__name__,
                        "token_type": AuthToken.token_type,
                        "message": e.args[0],
                    }
                )

        raise InvalidToken(
            {
                "detail": _("Given token not valid for any token type"),
                "messages": messages,
            }
        )

    def check_blacklist(self, token):
        if api_settings.JTI_CLAIM not in token:
            return

        jti = token[api_settings.JTI_CLAIM]
        SessionManagement.check_blacklist(jti)

    def get_user(self, validated_token):
        """
        Attempts to find and return a user using the given validated token.
        """
        try:
            user_id = validated_token[api_settings.USER_ID_CLAIM]
        except KeyError:
            raise InvalidToken(_("Token contained no recognizable user identification"))

        try:
            user = self.user_model.objects.get(**{api_settings.USER_ID_FIELD: user_id})
        except self.user_model.DoesNotExist:
            raise AuthenticationFailed(_("User not found"), code="user_not_found")

        if not user.is_active:
            raise AuthenticationFailed(_("User is inactive"), code="user_inactive")

        return user


class JWTStatelessUserAuthentication(JWTAuthentication):
    """
    An authentication plugin that authenticates requests through a JSON web
    token provided in a request header without performing a database lookup to obtain a user instance.
    """

    def get_user(self, validated_token):
        """
        Returns a stateless user object which is backed by the given validated
        token.
        """
        if api_settings.USER_ID_CLAIM not in validated_token:
            # The TokenUser class assumes tokens will have a recognizable user
            # identifier claim.
            raise InvalidToken(_("Token contained no recognizable user identification"))

        return api_settings.TOKEN_USER_CLASS(validated_token)


JWTTokenUserAuthentication = JWTStatelessUserAuthentication


def default_user_authentication_rule(user):
    # Prior to Django 1.10, inactive users could be authenticated with the
    # default `ModelBackend`.  As of Django 1.10, the `ModelBackend`
    # prevents inactive users from authenticating.  App designers can still
    # allow inactive users to authenticate by opting for the new
    # `AllowAllUsersModelBackend`.  However, we explicitly prevent inactive
    # users from authenticating to enforce a reasonable policy and provide
    # sensible backwards compatibility with older Django versions.
    return user is not None and user.is_active



import json
from datetime import timedelta
from typing import Optional, Type, Union

import jwt
from django.utils.translation import gettext_lazy as _
from jwt import InvalidAlgorithmError, InvalidTokenError, algorithms

from .exceptions import TokenBackendError
from .utils import format_lazy

try:
    from jwt import PyJWKClient, PyJWKClientError

    JWK_CLIENT_AVAILABLE = True
except ImportError:
    JWK_CLIENT_AVAILABLE = False

ALLOWED_ALGORITHMS = {
    "HS256",
    "HS384",
    "HS512",
    "RS256",
    "RS384",
    "RS512",
    "ES256",
    "ES384",
    "ES512",
}


class TokenBackend:
    def __init__(
        self,
        algorithm,
        signing_key=None,
        verifying_key="",
        audience=None,
        issuer=None,
        jwk_url: str = None,
        leeway: Union[float, int, timedelta] = None,
        json_encoder: Optional[Type[json.JSONEncoder]] = None,
    ):
        self._validate_algorithm(algorithm)

        self.algorithm = algorithm
        self.signing_key = signing_key
        self.verifying_key = verifying_key
        self.audience = audience
        self.issuer = issuer

        if JWK_CLIENT_AVAILABLE:
            self.jwks_client = PyJWKClient(jwk_url) if jwk_url else None
        else:
            self.jwks_client = None

        self.leeway = leeway
        self.json_encoder = json_encoder

    def _validate_algorithm(self, algorithm):
        """
        Ensure that the nominated algorithm is recognized, and that cryptography is installed for those
        algorithms that require it
        """
        if algorithm not in ALLOWED_ALGORITHMS:
            raise TokenBackendError(
                format_lazy(_("Unrecognized algorithm type '{}'"), algorithm)
            )

        if algorithm in algorithms.requires_cryptography and not algorithms.has_crypto:
            raise TokenBackendError(
                format_lazy(
                    _("You must have cryptography installed to use {}."), algorithm
                )
            )

    def get_leeway(self) -> timedelta:
        if self.leeway is None:
            return timedelta(seconds=0)
        elif isinstance(self.leeway, (int, float)):
            return timedelta(seconds=self.leeway)
        elif isinstance(self.leeway, timedelta):
            return self.leeway
        else:
            raise TokenBackendError(
                format_lazy(
                    _(
                        "Unrecognized type '{}', 'leeway' must be of type int, float or timedelta."
                    ),
                    type(self.leeway),
                )
            )

    def get_verifying_key(self, token):
        if self.algorithm.startswith("HS"):
            return self.signing_key

        if self.jwks_client:
            try:
                return self.jwks_client.get_signing_key_from_jwt(token).key
            except PyJWKClientError as ex:
                raise TokenBackendError(_("Token is invalid or expired")) from ex

        return self.verifying_key

    def encode(self, payload):
        """
        Returns an encoded token for the given payload dictionary.
        """
        jwt_payload = payload.copy()
        if self.audience is not None:
            jwt_payload["aud"] = self.audience
        if self.issuer is not None:
            jwt_payload["iss"] = self.issuer

        token = jwt.encode(
            jwt_payload,
            self.signing_key,
            algorithm=self.algorithm,
            json_encoder=self.json_encoder,
        )
        if isinstance(token, bytes):
            # For PyJWT <= 1.7.1
            return token.decode("utf-8")
        # For PyJWT >= 2.0.0a1
        return token

    def decode(self, token, verify=True):
        """
        Performs a validation of the given token and returns its payload
        dictionary.

        Raises a `TokenBackendError` if the token is malformed, if its
        signature check fails, or if its 'exp' claim indicates it has expired.
        """
        try:
            return jwt.decode(
                token,
                self.get_verifying_key(token),
                algorithms=[self.algorithm],
                audience=self.audience,
                issuer=self.issuer,
                leeway=self.get_leeway(),
                options={
                    "verify_aud": self.audience is not None,
                    "verify_signature": verify,
                },
            )
        except InvalidAlgorithmError as ex:
            raise TokenBackendError(_("Invalid algorithm specified")) from ex
        except InvalidTokenError as ex:
            raise TokenBackendError(_("Token is invalid or expired")) from ex



import time

from django.core.cache import cache
from django.utils import timezone
from rest_framework import status
from rest_framework.generics import ListAPIView
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response

from nilva_django_auth.models import UserSession
from nilva_django_auth.serializers import SessionSerializer
from nilva_django_auth.session_management import SessionManagement
from nilva_django_auth.simplejwt.authentication import JWTAuthentication
from nilva_django_auth.simplejwt.settings import api_settings


def cache_api_data(func):
    def inner(*args, **kwargs):
        func_self = args[0]

        cache_key = func_self.get_cache_key()
        start = time.time()
        cache_data = cache.get(cache_key)
        end = time.time()
        if cache_data is not None:
            return Response(cache_data)

        response = func(*args, **kwargs)
        if response.status_code == status.HTTP_200_OK:
            cache.set(cache_key, response.data, )
        return response

    return inner


class SessionListView(ListAPIView):
    """
    API view for listing user sessions.

    Returns a JSON response with a list of active sessions for the current user.
    """
    authentication_classes = [JWTAuthentication]
    permission_classes = [IsAuthenticated]
    serializer_class = SessionSerializer
    pagination_class = None

    def get_cache_key(self):
        return SessionManagement.get_session_list_cache_key(self.request.user.id)

    def get_queryset(self):
        """
        Get a list of all active sessions for the current user.

        Returns:
            QuerySet of active UserSession objects for the current user.
        """
        user_id = self.request.user.id

        # Get all active sessions for the user from the database
        return UserSession.objects.filter(user_id=user_id, is_active=True, exp__gte=timezone.now()).order_by(
            '-last_activity'
        )

    def get_serializer_context(self):
        """
        Add current_jti to the serializer context.
        """
        context = super().get_serializer_context()

        # Extract the JTI from the request's token
        token = self.request.auth
        if token:
            context['current_jti'] = token.get(api_settings.JTI_CLAIM)

        return context

    @cache_api_data
    def list(self, request, *args, **kwargs):
        return super().list(request, *args, **kwargs)



from django.urls import path

from nilva_django_auth.simplejwt.views import TokenObtainPairView, TokenRefreshView, TokenBlacklistView, \
    TokenBlacklistAllView, TokenVerifyView, TokenBlacklistSessionsView
from nilva_django_auth.views import SessionListView

app_name = 'nilva_django_auth'

urlpatterns = [
    path('login/', TokenObtainPairView.as_view(), name='login'),
    path('refresh/', TokenRefreshView.as_view(), name='refresh'),
    path('logout/', TokenBlacklistView.as_view(), name='logout'),
    path('logout-all/', TokenBlacklistAllView.as_view(), name='logout-all'),
    path('logout-sessions/', TokenBlacklistSessionsView.as_view(), name='logout-sessions'),
    path('verify/', TokenVerifyView.as_view(), name='token-verify'),
    path('sessions/', SessionListView.as_view(), name='sessions'),
]
